/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::Navigator
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <optional>
#include <string>
#include <vector>

// Eigen
#include <Eigen/Core>

// ArmarX
#include <ArmarXCore/core/Component.h>

#include <ArmarXGui/libraries/ArmarXGuiComponentPlugins/LightweightRemoteGuiComponentPlugin.h>

#include <RobotAPI/interface/units/PlatformUnitInterface.h>
#include <RobotAPI/libraries/ArmarXObjects/plugins/ObjectPoseClientPlugin.h>
#include <RobotAPI/libraries/RobotAPIComponentPlugins/RobotStateComponentPlugin.h>

// Navigation
#include "Navigation/libraries/core/StaticScene.h"
#include "Navigation/libraries/core/types.h"
#include <Navigation/components/Navigator/NavigatorInterface.h>
#include <Navigation/libraries/server/Navigator.h>
#include <Navigation/libraries/server/execution/PlatformUnitExecutor.h>

namespace armarx::nav::components
{

    /**
     * @defgroup Component-Navigator Navigator
     * @ingroup Navigation-Components
     * A description of the component Navigator.
     *
     * @class Navigator
     * @ingroup Component-Navigator
     * @brief Brief description of class Navigator.
     *
     * Detailed description of class Navigator.
     */
    class Navigator :
        virtual public armarx::Component,
        virtual public NavigatorInterface,
        virtual public armarx::ObjectPoseClientPluginUser,
        virtual public armarx::RobotStateComponentPluginUser,
        virtual public armarx::LightweightRemoteGuiComponentPluginUser
    {

    public:
        Navigator();
        ~Navigator() override;

        /// @see armarx::ManagedIceObject::getDefaultName()
        std::string getDefaultName() const override;

        std::string createConfig(const aron::data::AronDictPtr& stackConfig,
                                 const Ice::Current&) override;

        void moveTo(const std::vector<Eigen::Matrix4f>& waypoints,
                    const std::string& navigationMode,
                    const std::string& configId,
                    const Ice::Current&) override;

        void moveTowards(const Eigen::Vector3f& direction,
                         const std::string& navigationMode,
                         const std::string& configId,
                         const Ice::Current&) override;

        void pauseMovement(const Ice::Current&) override;

        void resumeMovement(const Ice::Current&) override;

    protected:
        /// @see PropertyUser::createPropertyDefinitions()
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

        /// @see armarx::ManagedIceObject::onInitComponent()
        void onInitComponent() override;

        /// @see armarx::ManagedIceObject::onConnectComponent()
        void onConnectComponent() override;

        /// @see armarx::ManagedIceObject::onDisconnectComponent()
        void onDisconnectComponent() override;

        /// @see armarx::ManagedIceObject::onExitComponent()
        void onExitComponent() override;

        void updateContext();

        core::StaticScene staticScene();

        VirtualRobot::RobotPtr getRobot();

        void createRemoteGuiTab();
        void RemoteGui_update() override;

    private:
        // TODO update context periodically

        PlatformUnitInterfacePrx platformUnit;

        core::Scene scene;
        server::PlatformUnitExecutor executor;
        std::map<std::string, server::Navigator> navigators;

        std::mutex propertiesMutex;

        struct RemoteGuiTab : armarx::RemoteGui::Client::Tab
        {

            struct
            {
                armarx::RemoteGui::Client::GroupBox group;

                armarx::RemoteGui::Client::FloatSpinBox targetPoseX;
                armarx::RemoteGui::Client::FloatSpinBox targetPoseY;
                armarx::RemoteGui::Client::FloatSpinBox targetPoseZ;
                armarx::RemoteGui::Client::FloatSpinBox targetPoseRoll;
                armarx::RemoteGui::Client::FloatSpinBox targetPosePitch;
                armarx::RemoteGui::Client::FloatSpinBox targetPoseYaw;
            } targetPoseGroup;

            struct
            {
                armarx::RemoteGui::Client::GroupBox group;

                armarx::RemoteGui::Client::Button moveToButton;
                armarx::RemoteGui::Client::Button pauseButton;
                armarx::RemoteGui::Client::Button continueButton;
            } controlGroup;

        };
        RemoteGuiTab tab;
    };
} // namespace armarx::nav::components

namespace armarx::nav
{
    using NavigatorComponent = components::Navigator;
}
